@@ -1,13 +1,11 @@
 /*
   +----------------------------------------------------------------------+
-  | PHP Version 5                                                        |
-  +----------------------------------------------------------------------+
-  | Copyright (c) 1997-2016 The PHP Group                                |
+  | Copyright (c) The PHP Group                                          |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
-  | http://www.php.net/license/3_01.txt                                  |
+  | https://www.php.net/license/3_01.txt                                 |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
@@ -16,19 +14,17 @@
   +----------------------------------------------------------------------+
 */
 
-/* $Id$ */
-
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
 #include "php.h"
 #include "php_bz2.h"
+#include "bz2_arginfo.h"
 
-#if HAVE_BZ2
+#ifdef HAVE_BZ2
 
 /* PHP Includes */
-#include "ext/standard/file.h"
 #include "ext/standard/info.h"
 #include "ext/standard/php_string.h"
 #include "main/php_network.h"
@@ -44,83 +40,17 @@
 static PHP_MINIT_FUNCTION(bz2);
 static PHP_MSHUTDOWN_FUNCTION(bz2);
 static PHP_MINFO_FUNCTION(bz2);
-static PHP_FUNCTION(bzopen);
-static PHP_FUNCTION(bzread);
-static PHP_FUNCTION(bzerrno);
-static PHP_FUNCTION(bzerrstr);
-static PHP_FUNCTION(bzerror);
-static PHP_FUNCTION(bzcompress);
-static PHP_FUNCTION(bzdecompress);
-
-/* {{{ arginfo */
-ZEND_BEGIN_ARG_INFO_EX(arginfo_bzread, 0, 0, 1)
-	ZEND_ARG_INFO(0, bz)
-	ZEND_ARG_INFO(0, length)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_bzopen, 0)
-	ZEND_ARG_INFO(0, file)
-	ZEND_ARG_INFO(0, mode)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_bzerrno, 0)
-	ZEND_ARG_INFO(0, bz)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_bzerrstr, 0)
-	ZEND_ARG_INFO(0, bz)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_bzerror, 0)
-	ZEND_ARG_INFO(0, bz)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_bzcompress, 0, 0, 2)
-	ZEND_ARG_INFO(0, source)
-	ZEND_ARG_INFO(0, blocksize)
-	ZEND_ARG_INFO(0, workfactor)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_bzdecompress, 0, 0, 1)
-	ZEND_ARG_INFO(0, source)
-	ZEND_ARG_INFO(0, small)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_bzwrite, 0, 0, 2)
-	ZEND_ARG_INFO(0, fp)
-	ZEND_ARG_INFO(0, str)
-	ZEND_ARG_INFO(0, length)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_bzflush, 0)
-	ZEND_ARG_INFO(0, fp)
-ZEND_END_ARG_INFO()
-/* }}} */
-
-static const zend_function_entry bz2_functions[] = {
-	PHP_FE(bzopen,       arginfo_bzopen)
-	PHP_FE(bzread,       arginfo_bzread)
-	PHP_FALIAS(bzwrite,   fwrite,		arginfo_bzwrite)
-	PHP_FALIAS(bzflush,   fflush,		arginfo_bzflush)
-	PHP_FALIAS(bzclose,   fclose,		arginfo_bzflush)
-	PHP_FE(bzerrno,      arginfo_bzerrno)
-	PHP_FE(bzerrstr,     arginfo_bzerrstr)
-	PHP_FE(bzerror,      arginfo_bzerror)
-	PHP_FE(bzcompress,   arginfo_bzcompress)
-	PHP_FE(bzdecompress, arginfo_bzdecompress)
-	PHP_FE_END
-};
 
 zend_module_entry bz2_module_entry = {
 	STANDARD_MODULE_HEADER,
 	"bz2",
-	bz2_functions,
+	ext_functions,
 	PHP_MINIT(bz2),
 	PHP_MSHUTDOWN(bz2),
 	NULL,
 	NULL,
 	PHP_MINFO(bz2),
-	NO_VERSION_YET,
+	PHP_BZ2_VERSION,
 	STANDARD_MODULE_PROPERTIES
 };
 
@@ -135,32 +65,65 @@ struct php_bz2_stream_data_t {
 
 /* {{{ BZip2 stream implementation */
 
-static size_t php_bz2iop_read(php_stream *stream, char *buf, size_t count TSRMLS_DC)
+static ssize_t php_bz2iop_read(php_stream *stream, char *buf, size_t count)
 {
-	struct php_bz2_stream_data_t *self = (struct php_bz2_stream_data_t *) stream->abstract;
-	int bz2_ret;
+	struct php_bz2_stream_data_t *self = (struct php_bz2_stream_data_t *)stream->abstract;
+	size_t ret = 0;
 
-	bz2_ret = BZ2_bzread(self->bz_file, buf, count);
+	do {
+		int just_read;
+		size_t remain = count - ret;
+		int to_read = (int)(remain <= INT_MAX ? remain : INT_MAX);
 
-	if (bz2_ret < 0) {
-		stream->eof = 1;
-		return -1;
-	}
-	if (bz2_ret == 0) {
-		stream->eof = 1;
-	}
+		just_read = BZ2_bzread(self->bz_file, buf, to_read);
 
-	return (size_t)bz2_ret;
+		if (just_read < 1) {
+			/* it is not safe to keep reading after an error, see #72613 */
+			stream->eof = 1;
+			if (just_read < 0) {
+				if (ret) {
+					return ret;
+				}
+				return -1;
+			}
+			break;
+		}
+
+		ret += just_read;
+	} while (ret < count);
+
+	return ret;
 }
 
-static size_t php_bz2iop_write(php_stream *stream, const char *buf, size_t count TSRMLS_DC)
+static ssize_t php_bz2iop_write(php_stream *stream, const char *buf, size_t count)
 {
-	struct php_bz2_stream_data_t *self = (struct php_bz2_stream_data_t *) stream->abstract;
+	ssize_t wrote = 0;
+	struct php_bz2_stream_data_t *self = (struct php_bz2_stream_data_t *)stream->abstract;
 
-	return BZ2_bzwrite(self->bz_file, (char*)buf, count);
+	do {
+		int just_wrote;
+		size_t remain = count - wrote;
+		int to_write = (int)(remain <= INT_MAX ? remain : INT_MAX);
+
+		just_wrote = BZ2_bzwrite(self->bz_file, (char*)buf, to_write);
+		if (just_wrote < 0) {
+			if (wrote == 0) {
+				return just_wrote;
+			}
+			return wrote;
+		}
+		if (just_wrote == 0) {
+			break;
+		}
+
+		wrote += just_wrote;
+
+	} while (wrote < count);
+
+	return wrote;
 }
 
-static int php_bz2iop_close(php_stream *stream, int close_handle TSRMLS_DC)
+static int php_bz2iop_close(php_stream *stream, int close_handle)
 {
 	struct php_bz2_stream_data_t *self = (struct php_bz2_stream_data_t *)stream->abstract;
 	int ret = EOF;
@@ -178,14 +141,14 @@ static int php_bz2iop_close(php_stream *stream, int close_handle TSRMLS_DC)
 	return ret;
 }
 
-static int php_bz2iop_flush(php_stream *stream TSRMLS_DC)
+static int php_bz2iop_flush(php_stream *stream)
 {
 	struct php_bz2_stream_data_t *self = (struct php_bz2_stream_data_t *)stream->abstract;
 	return BZ2_bzflush(self->bz_file);
 }
 /* }}} */
 
-php_stream_ops php_stream_bz2io_ops = {
+const php_stream_ops php_stream_bz2io_ops = {
 	php_bz2iop_write, php_bz2iop_read,
 	php_bz2iop_close, php_bz2iop_flush,
 	"BZip2",
@@ -196,14 +159,17 @@ php_stream_ops php_stream_bz2io_ops = {
 };
 
 /* {{{ Bzip2 stream openers */
-PHP_BZ2_API php_stream *_php_stream_bz2open_from_BZFILE(BZFILE *bz, 
-														const char *mode, php_stream *innerstream STREAMS_DC TSRMLS_DC)
+PHP_BZ2_API php_stream *_php_stream_bz2open_from_BZFILE(BZFILE *bz,
+														const char *mode, php_stream *innerstream STREAMS_DC)
 {
 	struct php_bz2_stream_data_t *self;
 
 	self = emalloc(sizeof(*self));
 
 	self->stream = innerstream;
+	if (innerstream) {
+		GC_ADDREF(innerstream->res);
+	}
 	self->bz_file = bz;
 
 	return php_stream_alloc_rel(&php_stream_bz2io_ops, self, 0, mode);
@@ -213,8 +179,8 @@ PHP_BZ2_API php_stream *_php_stream_bz2open(php_stream_wrapper *wrapper,
 											const char *path,
 											const char *mode,
 											int options,
-											char **opened_path,
-											php_stream_context *context STREAMS_DC TSRMLS_DC)
+											zend_string **opened_path,
+											php_stream_context *context STREAMS_DC)
 {
 	php_stream *retstream = NULL, *stream = NULL;
 	char *path_copy = NULL;
@@ -228,12 +194,12 @@ PHP_BZ2_API php_stream *_php_stream_bz2open(php_stream_wrapper *wrapper,
 	}
 
 #ifdef VIRTUAL_DIR
-	virtual_filepath_ex(path, &path_copy, NULL TSRMLS_CC);
+	virtual_filepath_ex(path, &path_copy, NULL);
 #else
-	path_copy = path;
+	path_copy = (char *)path;
 #endif
 
-	if (php_check_open_basedir(path_copy TSRMLS_CC)) {
+	if (php_check_open_basedir(path_copy)) {
 #ifdef VIRTUAL_DIR
 		efree(path_copy);
 #endif
@@ -244,12 +210,12 @@ PHP_BZ2_API php_stream *_php_stream_bz2open(php_stream_wrapper *wrapper,
 	bz_file = BZ2_bzopen(path_copy, mode);
 
 	if (opened_path && bz_file) {
-		*opened_path = estrdup(path_copy);
+		*opened_path = zend_string_init(path_copy, strlen(path_copy), 0);
 	}
+
 #ifdef VIRTUAL_DIR
 	efree(path_copy);
 #endif
-	path_copy = NULL;
 
 	if (bz_file == NULL) {
 		/* that didn't work, so try and get something from the network/wrapper */
@@ -258,7 +224,7 @@ PHP_BZ2_API php_stream *_php_stream_bz2open(php_stream_wrapper *wrapper,
 		if (stream) {
 			php_socket_t fd;
 			if (SUCCESS == php_stream_cast(stream, PHP_STREAM_AS_FD, (void **) &fd, REPORT_ERRORS)) {
-				bz_file = BZ2_bzdopen(fd, mode);
+				bz_file = BZ2_bzdopen((int)fd, mode);
 			}
 		}
 
@@ -266,12 +232,12 @@ PHP_BZ2_API php_stream *_php_stream_bz2open(php_stream_wrapper *wrapper,
 		 * failed.
 		 */
 		if (opened_path && !bz_file && mode[0] == 'w') {
-			VCWD_UNLINK(*opened_path);
+			VCWD_UNLINK(ZSTR_VAL(*opened_path));
 		}
 	}
 
 	if (bz_file) {
-		retstream = _php_stream_bz2open_from_BZFILE(bz_file, mode, stream STREAMS_REL_CC TSRMLS_CC);
+		retstream = _php_stream_bz2open_from_BZFILE(bz_file, mode, stream STREAMS_REL_CC);
 		if (retstream) {
 			return retstream;
 		}
@@ -288,7 +254,7 @@ PHP_BZ2_API php_stream *_php_stream_bz2open(php_stream_wrapper *wrapper,
 
 /* }}} */
 
-static php_stream_wrapper_ops bzip2_stream_wops = {
+static const php_stream_wrapper_ops bzip2_stream_wops = {
 	_php_stream_bz2open,
 	NULL, /* close */
 	NULL, /* fstat */
@@ -298,10 +264,11 @@ static php_stream_wrapper_ops bzip2_stream_wops = {
 	NULL, /* unlink */
 	NULL, /* rename */
 	NULL, /* mkdir */
-	NULL  /* rmdir */
+	NULL, /* rmdir */
+	NULL
 };
 
-static php_stream_wrapper php_stream_bzip2_wrapper = {
+static const php_stream_wrapper php_stream_bzip2_wrapper = {
 	&bzip2_stream_wops,
 	NULL,
 	0 /* is_url */
@@ -311,15 +278,15 @@ static void php_bz2_error(INTERNAL_FUNCTION_PARAMETERS, int);
 
 static PHP_MINIT_FUNCTION(bz2)
 {
-	php_register_url_stream_wrapper("compress.bzip2", &php_stream_bzip2_wrapper TSRMLS_CC);
-	php_stream_filter_register_factory("bzip2.*", &php_bz2_filter_factory TSRMLS_CC);
+	php_register_url_stream_wrapper("compress.bzip2", &php_stream_bzip2_wrapper);
+	php_stream_filter_register_factory("bzip2.*", &php_bz2_filter_factory);
 	return SUCCESS;
 }
 
 static PHP_MSHUTDOWN_FUNCTION(bz2)
 {
-	php_unregister_url_stream_wrapper("compress.bzip2" TSRMLS_CC);
-	php_stream_filter_unregister_factory("bzip2.*" TSRMLS_CC);
+	php_unregister_url_stream_wrapper("compress.bzip2");
+	php_stream_filter_unregister_factory("bzip2.*");
 
 	return SUCCESS;
 }
@@ -334,88 +301,78 @@ static PHP_MINFO_FUNCTION(bz2)
 	php_info_print_table_end();
 }
 
-/* {{{ proto string bzread(resource bz[, int length])
-   Reads up to length bytes from a BZip2 stream, or 1024 bytes if length is not specified */
-static PHP_FUNCTION(bzread)
+/* {{{ Reads up to length bytes from a BZip2 stream, or 1024 bytes if length is not specified */
+PHP_FUNCTION(bzread)
 {
 	zval *bz;
-	long len = 1024;
+	zend_long len = 1024;
 	php_stream *stream;
+	zend_string *data;
 
-	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r|l", &bz, &len)) {
-		RETURN_FALSE;
+	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS(), "r|l", &bz, &len)) {
+		RETURN_THROWS();
 	}
 
-	php_stream_from_zval(stream, &bz);
+	php_stream_from_zval(stream, bz);
 
-	if ((len + 1) < 1) {
-		php_error_docref(NULL TSRMLS_CC, E_WARNING, "length may not be negative");
-		RETURN_FALSE;
+	if (len  < 0) {
+		zend_argument_value_error(2, "must be greater than or equal to 0");
+		RETURN_THROWS();
 	}
 
-	Z_STRVAL_P(return_value) = emalloc(len + 1);
-	Z_STRLEN_P(return_value) = php_stream_read(stream, Z_STRVAL_P(return_value), len);
-
-	if (Z_STRLEN_P(return_value) < 0) {
-		efree(Z_STRVAL_P(return_value));
-		php_error_docref(NULL TSRMLS_CC, E_WARNING, "could not read valid bz2 data from stream");
+	data = php_stream_read_to_str(stream, len);
+	if (!data) {
 		RETURN_FALSE;
 	}
-
-	Z_STRVAL_P(return_value)[Z_STRLEN_P(return_value)] = 0;
-	Z_TYPE_P(return_value) = IS_STRING;
+	RETURN_STR(data);
 }
 /* }}} */
 
-/* {{{ proto resource bzopen(string|int file|fp, string mode)
-   Opens a new BZip2 stream */
-static PHP_FUNCTION(bzopen)
+/* {{{ Opens a new BZip2 stream */
+PHP_FUNCTION(bzopen)
 {
-	zval    **file;   /* The file to open */
+	zval     *file;   /* The file to open */
 	char     *mode;   /* The mode to open the stream with */
-	int      mode_len;
+	size_t      mode_len;
 
 	BZFILE   *bz;     /* The compressed file stream */
 	php_stream *stream = NULL;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "Zs", &file, &mode, &mode_len) == FAILURE) {
-		return;
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "zs", &file, &mode, &mode_len) == FAILURE) {
+		RETURN_THROWS();
 	}
 
 	if (mode_len != 1 || (mode[0] != 'r' && mode[0] != 'w')) {
-		php_error_docref(NULL TSRMLS_CC, E_WARNING, "'%s' is not a valid mode for bzopen(). Only 'w' and 'r' are supported.", mode);
-		RETURN_FALSE;
+		zend_argument_value_error(2, "must be either \"r\" or \"w\"");
+		RETURN_THROWS();
 	}
 
 	/* If it's not a resource its a string containing the filename to open */
-	if (Z_TYPE_PP(file) == IS_STRING) {
-		if (Z_STRLEN_PP(file) == 0) {
-			php_error_docref(NULL TSRMLS_CC, E_WARNING, "filename cannot be empty");
-			RETURN_FALSE;
+	if (Z_TYPE_P(file) == IS_STRING) {
+		if (Z_STRLEN_P(file) == 0) {
+			zend_argument_value_error(1, "cannot be empty");
+			RETURN_THROWS();
 		}
 
-		if (CHECK_ZVAL_NULL_PATH(*file)) {
-			RETURN_FALSE;
+		if (CHECK_ZVAL_NULL_PATH(file)) {
+			zend_argument_type_error(1, "must not contain null bytes");
+			RETURN_THROWS();
 		}
 
-		stream = php_stream_bz2open(NULL,
-									Z_STRVAL_PP(file),
-									mode,
-									REPORT_ERRORS,
-									NULL);
-	} else if (Z_TYPE_PP(file) == IS_RESOURCE) {
+		stream = php_stream_bz2open(NULL, Z_STRVAL_P(file), mode, REPORT_ERRORS, NULL);
+	} else if (Z_TYPE_P(file) == IS_RESOURCE) {
 		/* If it is a resource, than its a stream resource */
 		php_socket_t fd;
-		int stream_mode_len;
+		size_t stream_mode_len;
 
 		php_stream_from_zval(stream, file);
 		stream_mode_len = strlen(stream->mode);
 
 		if (stream_mode_len != 1 && !(stream_mode_len == 2 && memchr(stream->mode, 'b', 2))) {
-			php_error_docref(NULL TSRMLS_CC, E_WARNING, "cannot use stream opened in mode '%s'", stream->mode);
+			php_error_docref(NULL, E_WARNING, "Cannot use stream opened in mode '%s'", stream->mode);
 			RETURN_FALSE;
 		} else if (stream_mode_len == 1 && stream->mode[0] != 'r' && stream->mode[0] != 'w' && stream->mode[0] != 'a' && stream->mode[0] != 'x') {
-			php_error_docref(NULL TSRMLS_CC, E_WARNING, "cannot use stream opened in mode '%s'", stream->mode);
+			php_error_docref(NULL, E_WARNING, "Cannot use stream opened in mode '%s'", stream->mode);
 			RETURN_FALSE;
 		}
 
@@ -423,7 +380,7 @@ static PHP_FUNCTION(bzopen)
 			case 'r':
 				/* only "r" and "rb" are supported */
 				if (stream->mode[0] != mode[0] && !(stream_mode_len == 2 && stream->mode[1] != mode[0])) {
-					php_error_docref(NULL TSRMLS_CC, E_WARNING, "cannot read from a stream opened in write only mode");
+					php_error_docref(NULL, E_WARNING, "Cannot read from a stream opened in write only mode");
 					RETURN_FALSE;
 				}
 				break;
@@ -432,25 +389,23 @@ static PHP_FUNCTION(bzopen)
 				if (stream->mode[0] != mode[0] && !(stream_mode_len == 2 && stream->mode[1] != mode[0])
 					&& stream->mode[0] != 'a' && !(stream_mode_len == 2 && stream->mode[1] != 'a')
 					&& stream->mode[0] != 'x' && !(stream_mode_len == 2 && stream->mode[1] != 'x')) {
-					php_error_docref(NULL TSRMLS_CC, E_WARNING, "cannot write to a stream opened in read only mode");
+					php_error_docref(NULL, E_WARNING, "cannot write to a stream opened in read only mode");
 					RETURN_FALSE;
 				}
 				break;
-			default:
-				/* not reachable */
-				break;
+			EMPTY_SWITCH_DEFAULT_CASE();
 		}
 
 		if (FAILURE == php_stream_cast(stream, PHP_STREAM_AS_FD, (void *) &fd, REPORT_ERRORS)) {
 			RETURN_FALSE;
 		}
 
-		bz = BZ2_bzdopen(fd, mode);
+		bz = BZ2_bzdopen((int)fd, mode);
 
 		stream = php_stream_bz2open_from_BZFILE(bz, mode, stream);
 	} else {
-		php_error_docref(NULL TSRMLS_CC, E_WARNING, "first parameter has to be string or file-resource");
-		RETURN_FALSE;
+		zend_argument_type_error(1, "must be of type string or file-resource, %s given", zend_zval_type_name(file));
+		RETURN_THROWS();
 	}
 
 	if (stream) {
@@ -461,59 +416,53 @@ static PHP_FUNCTION(bzopen)
 }
 /* }}} */
 
-/* {{{ proto int bzerrno(resource bz)
-   Returns the error number */
-static PHP_FUNCTION(bzerrno)
+/* {{{ Returns the error number */
+PHP_FUNCTION(bzerrno)
 {
 	php_bz2_error(INTERNAL_FUNCTION_PARAM_PASSTHRU, PHP_BZ_ERRNO);
 }
 /* }}} */
 
-/* {{{ proto string bzerrstr(resource bz)
-   Returns the error string */
-static PHP_FUNCTION(bzerrstr)
+/* {{{ Returns the error string */
+PHP_FUNCTION(bzerrstr)
 {
 	php_bz2_error(INTERNAL_FUNCTION_PARAM_PASSTHRU, PHP_BZ_ERRSTR);
 }
 /* }}} */
 
-/* {{{ proto array bzerror(resource bz)
-   Returns the error number and error string in an associative array */
-static PHP_FUNCTION(bzerror)
+/* {{{ Returns the error number and error string in an associative array */
+PHP_FUNCTION(bzerror)
 {
 	php_bz2_error(INTERNAL_FUNCTION_PARAM_PASSTHRU, PHP_BZ_ERRBOTH);
 }
 /* }}} */
 
-/* {{{ proto string bzcompress(string source [, int blocksize100k [, int workfactor]])
-   Compresses a string into BZip2 encoded data */
-static PHP_FUNCTION(bzcompress)
+/* {{{ Compresses a string into BZip2 encoded data */
+PHP_FUNCTION(bzcompress)
 {
 	char             *source;          /* Source data to compress */
-	long              zblock_size = 0; /* Optional block size to use */
-	long              zwork_factor = 0;/* Optional work factor to use */
-	char             *dest = NULL;     /* Destination to place the compressed data into */
+	zend_long              zblock_size = 0; /* Optional block size to use */
+	zend_long              zwork_factor = 0;/* Optional work factor to use */
+	zend_string      *dest = NULL;     /* Destination to place the compressed data into */
 	int               error,           /* Error Container */
 					  block_size  = 4, /* Block size for compression algorithm */
 					  work_factor = 0, /* Work factor for compression algorithm */
-					  argc;            /* Argument count */
-	int               source_len;      /* Length of the source data */
+					  argc = ZEND_NUM_ARGS(); /* Argument count */
+	size_t               source_len;      /* Length of the source data */
 	unsigned int      dest_len;        /* Length of the destination buffer */
 
-	argc = ZEND_NUM_ARGS();
-
-	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|ll", &source, &source_len, &zblock_size, &zwork_factor) == FAILURE) {
-		return;
+	if (zend_parse_parameters(argc, "s|ll", &source, &source_len, &zblock_size, &zwork_factor) == FAILURE) {
+		RETURN_THROWS();
 	}
 
 	/* Assign them to easy to use variables, dest_len is initially the length of the data
 	   + .01 x length of data + 600 which is the largest size the results of the compression
 	   could possibly be, at least that's what the libbz2 docs say (thanks to jeremy@nirvani.net
 	   for pointing this out).  */
-	dest_len   = (unsigned int) (source_len + (0.01 * source_len) + 600);
+	dest_len = (unsigned int) (source_len + (0.01 * source_len) + 600);
 
 	/* Allocate the destination buffer */
-	dest = emalloc(dest_len + 1);
+	dest = zend_string_alloc(dest_len, 0);
 
 	/* Handle the optional arguments */
 	if (argc > 1) {
@@ -524,42 +473,43 @@ static PHP_FUNCTION(bzcompress)
 		work_factor = zwork_factor;
 	}
 
-	error = BZ2_bzBuffToBuffCompress(dest, &dest_len, source, source_len, block_size, 0, work_factor);
+	error = BZ2_bzBuffToBuffCompress(ZSTR_VAL(dest), &dest_len, source, source_len, block_size, 0, work_factor);
 	if (error != BZ_OK) {
-		efree(dest);
+		zend_string_efree(dest);
 		RETURN_LONG(error);
 	} else {
 		/* Copy the buffer, we have perhaps allocate a lot more than we need,
 		   so we erealloc() the buffer to the proper size */
-		dest = erealloc(dest, dest_len + 1);
-		dest[dest_len] = 0;
-		RETURN_STRINGL(dest, dest_len, 0);
+		ZSTR_LEN(dest) = dest_len;
+		ZSTR_VAL(dest)[ZSTR_LEN(dest)] = '\0';
+		RETURN_NEW_STR(dest);
 	}
 }
 /* }}} */
 
-/* {{{ proto string bzdecompress(string source [, int small])
-   Decompresses BZip2 compressed data */
-static PHP_FUNCTION(bzdecompress)
+/* {{{ Decompresses BZip2 compressed data */
+PHP_FUNCTION(bzdecompress)
 {
-	char *source, *dest;
-	int source_len, error;
-	long small = 0;
-#if defined(PHP_WIN32)
+	char *source;
+	zend_string *dest;
+	size_t source_len;
+	int error;
+	bool small = 0;
+#ifdef PHP_WIN32
 	unsigned __int64 size = 0;
 #else
 	unsigned long long size = 0;
 #endif
 	bz_stream bzs;
 
-	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|l", &source, &source_len, &small)) {
-		RETURN_FALSE;
+	if (FAILURE == zend_parse_parameters(ZEND_NUM_ARGS(), "s|b", &source, &source_len, &small)) {
+		RETURN_THROWS();
 	}
 
 	bzs.bzalloc = NULL;
 	bzs.bzfree = NULL;
 
-	if (BZ2_bzDecompressInit(&bzs, 0, small) != BZ_OK) {
+	if (BZ2_bzDecompressInit(&bzs, 0, (int)small) != BZ_OK) {
 		RETURN_FALSE;
 	}
 
@@ -567,24 +517,41 @@ static PHP_FUNCTION(bzdecompress)
 	bzs.avail_in = source_len;
 
 	/* in most cases bz2 offers at least 2:1 compression, so we use that as our base */
+	dest = zend_string_safe_alloc(source_len, 2, 1, 0);
 	bzs.avail_out = source_len * 2;
-	bzs.next_out = dest = emalloc(bzs.avail_out + 1);
+	bzs.next_out = ZSTR_VAL(dest);
 
 	while ((error = BZ2_bzDecompress(&bzs)) == BZ_OK && bzs.avail_in > 0) {
 		/* compression is better then 2:1, need to allocate more memory */
 		bzs.avail_out = source_len;
 		size = (bzs.total_out_hi32 * (unsigned int) -1) + bzs.total_out_lo32;
-		dest = safe_erealloc(dest, 1, bzs.avail_out+1, (size_t) size );
-		bzs.next_out = dest + size;
+#ifndef ZEND_ENABLE_ZVAL_LONG64
+		if (size > SIZE_MAX) {
+			/* no reason to continue if we're going to drop it anyway */
+			break;
+		}
+#endif
+		dest = zend_string_safe_realloc(dest, 1, bzs.avail_out+1, (size_t) size, 0);
+		bzs.next_out = ZSTR_VAL(dest) + size;
 	}
 
 	if (error == BZ_STREAM_END || error == BZ_OK) {
 		size = (bzs.total_out_hi32 * (unsigned int) -1) + bzs.total_out_lo32;
-		dest = safe_erealloc(dest, 1, (size_t) size, 1);
-		dest[size] = '\0';
-		RETVAL_STRINGL(dest, (int) size, 0);
+#ifndef ZEND_ENABLE_ZVAL_LONG64
+		if (UNEXPECTED(size > SIZE_MAX)) {
+			php_error_docref(NULL, E_WARNING, "Decompressed size too big, max is %zd", SIZE_MAX);
+			zend_string_efree(dest);
+			RETVAL_LONG(BZ_MEM_ERROR);
+		} else
+#endif
+		{
+			dest = zend_string_safe_realloc(dest, 1, (size_t)size, 1, 0);
+			ZSTR_LEN(dest) = (size_t)size;
+			ZSTR_VAL(dest)[(size_t)size] = '\0';
+			RETVAL_STR(dest);
+		}
 	} else { /* real error */
-		efree(dest);
+		zend_string_efree(dest);
 		RETVAL_LONG(error);
 	}
 
@@ -602,14 +569,15 @@ static void php_bz2_error(INTERNAL_FUNCTION_PARAMETERS, int opt)
 	int           errnum;  /* Error number */
 	struct php_bz2_stream_data_t *self;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &bzp) == FAILURE) {
-		return;
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "r", &bzp) == FAILURE) {
+		RETURN_THROWS();
 	}
 
-	php_stream_from_zval(stream, &bzp);
+	php_stream_from_zval(stream, bzp);
 
 	if (!php_stream_is(stream, PHP_STREAM_IS_BZIP2)) {
-		RETURN_FALSE;
+		zend_argument_type_error(1, "must be a bz2 stream");
+		RETURN_THROWS();
 	}
 
 	self = (struct php_bz2_stream_data_t *) stream->abstract;
@@ -623,25 +591,16 @@ static void php_bz2_error(INTERNAL_FUNCTION_PARAMETERS, int opt)
 			RETURN_LONG(errnum);
 			break;
 		case PHP_BZ_ERRSTR:
-			RETURN_STRING((char*)errstr, 1);
+			RETURN_STRING((char*)errstr);
 			break;
 		case PHP_BZ_ERRBOTH:
 			array_init(return_value);
 
 			add_assoc_long  (return_value, "errno",  errnum);
-			add_assoc_string(return_value, "errstr", (char*)errstr, 1);
+			add_assoc_string(return_value, "errstr", (char*)errstr);
 			break;
 	}
 }
 /* }}} */
 
 #endif
-
-/*
- * Local variables:
- * tab-width: 4
- * c-basic-offset: 4
- * End:
- * vim600: fdm=marker
- * vim: noet sw=4 ts=4
- */
